import { isObject, selectRandom, mergeDeep, nodeFromPath } from './util';

export default function Baba(rawSpec) {
	const grammarSpec = mergeDeep({
		// Defaults go here
	}, rawSpec);
	const grammarSpecOrig = JSON.parse(JSON.stringify(grammarSpec));

	const parseTransforms = (obj, leafPath=[], target={}) => {
		const path = leafPath.join('.');

		if (isObject(obj)) {
			// Flatten subobjects with full node path
			for (const [key, value] of Object.entries(obj)) {
				parseTransforms(value, leafPath.concat([key]), target);
			}
		}
		else {
			target[path] = new Function('$in', '$spec', '$helpers', obj).bind({});
		}

		return target;
	};

	const helpers = {
		re: (str, rules) => {
			// Loop through an array of regexps and bail when the first one 
			// succeeds
			let ret;
			rules.some(filter => {
				if (str.match(filter[0])) {
					ret = str.replace(filter[0], filter[1]);
					return true;
				}
				return false;
			});
			return ret;
		},
	};

	// Flatten grammar tree and handle leaf nodes (split, expose, etc.)
	const parseString = input => {
		let ret = input;

		// Replace lists
		ret = ret.replace(/\[(.*?)\]/g, (m, match) => {
			const matches = match.split('|');
			if (matches.length === 1) {
				// Lists with only one item will be included half the time
				matches.push('');
			}
			return selectRandom(matches);
		});

		// Replace references
		ret = ret.replace(/\{(\$([a-zA-Z0-9.\-_]+)\s*|([a-zA-Z0-9.\-_]+))\s*([a-zA-Z0-9.\-_ ]*)\}/g, (m, _, variable, ref, tf) => {
			let transforms = (tf && tf.split(/\s+/) || []);
			let value;

			transforms = transforms.map(item => {
				const transformFunc = grammarSpec.transform.function[item];
				if (!transformFunc) {
					console.error('Transform function missing: "%s"', item);
					return noop => noop;
				}
				return transformFunc;
			});

			try {
				// Try to load reference from grammar, use stored variable if 
				// available
				value = variable && grammarSpec.variable[variable]
					? grammarSpec.variable[variable]
					: grammarSpec.ruleFromPath(ref)();
			}
			catch (e) {
				// Not found in grammar
				console.error('Reference missing: "%s"', ref);
				value = ref;
			}

			transforms.forEach(transform =>
				value = transform(value, grammarSpec, helpers));

			return value;
		});

		return ret;
	};

	const parseRules = (node, path=[]) => {
		let ret = node;
		if (isObject(node)) {
			for (const [key, value] of Object.entries(node)) {
				ret[key] = parseRules(value, path.concat(key));
			}
		}
		else {
			ret = Array.isArray(ret) ? ret : [ret.trim()];

			const values = ret.reduce((arr, node) => {
				let values = node; // Plain string node
				let weight = 1;
				if (isObject(node)) {
					// Duplicated weighted nodes to increase probability of the 
					// nodes to be selected
					if (node.weight) {
						weight = node.weight;
					}

					if (node.hasOwnProperty('split') && node.split) {
						if (node.text) {
							// Delimited string that should be split with optional 
							// custom delimiter
							const delimiter = typeof(node.split) == 'boolean'
								? grammarSpec.meta.delimiter.split
								: node.split;
							values = node.text.split(delimiter);
						}
						else {
							// Split shorthand (single property), split with 
							// default delimiter
							values = node.split.split(grammarSpec.meta.delimiter.split);
						}
						values = values.map(item => item.trim());
					}
					else {
						// Plain string node (object)
						values = node.text;
					}
				}

				return arr.concat(new Array(weight).fill(values));
			}, []);

			ret = () => parseString(selectRandom(values));
		}
		return ret;
	};

	grammarSpec.transform.function = parseTransforms(grammarSpec.transform.function);
	grammarSpec.rule = parseRules(grammarSpec.rule);
	grammarSpec.ruleFromPath = path => {
		// Always returns a callable rule
		const node = nodeFromPath(grammarSpec.rule, path);
		if (!(node instanceof Function)) {
			// Randomly select nodes until we reach a callable rule
			let randomKey;
			let randomNode = node;
			do {
				randomKey = selectRandom(Object.keys(randomNode));
				randomNode = randomNode[randomKey];
				if (randomNode instanceof Function) {
					return randomNode;
				}
			} while (typeof randomNode !== 'undefined');
			return () => console.error('No rule found at path "%s"', path);
		}
		return node;
	};

	grammarSpec.setVariables = variables => {
		for (const [key, value] of Object.entries(mergeDeep(grammarSpecOrig.variable, variables || {}))) {
			grammarSpec.variable[key] = parseString(value);
		}
	};

	grammarSpec.setVariables(); // Set default variables

	return grammarSpec;
}
